﻿Imports System.Globalization
Imports System.Globalization.CultureInfo
Imports System.Management
Imports System.IO
Imports System.Text
Imports System.Runtime.InteropServices

Public Class USBViewer


    Public Sub New()
        InitializeComponent()
        getDevices()
    End Sub
    Private Sub USBViewer_KeyDown(ByVal sender As Object, ByVal e As System.Windows.Forms.KeyEventArgs) Handles Me.KeyDown
        lstUSBDisks.Items.Clear()
        getDevices()
    End Sub
    Private Sub getDevices()
        For Each device As USBDevice In GetConnectedDevices()
            Dim lvi As New ListViewItem

            lvi.Text = device.Manufacturer
            lvi.SubItems.Add(device.Product)
            'lvi.SubItems.Add(device.InstanceID)
            Dim Instance As String = device.InstanceID
            Dim split As String() = Instance.Split(New [Char]() {"\"c, "_"c, "&"c, CChar(vbTab)})
            lvi.SubItems.Add(split(2))
            lvi.SubItems.Add(split(4))
            lvi.SubItems.Add(device.SerialNumber)
            lstUSBDisks.Items.Add(lvi)
        Next
    End Sub

#Region "API Region"

    ' ********************** Constants ************************

    Const GENERIC_WRITE As Integer = &H40000000
    Const FILE_SHARE_READ As Integer = &H1
    Const FILE_SHARE_WRITE As Integer = &H2
    Const OPEN_EXISTING As Integer = &H3
    Const INVALID_HANDLE_VALUE As Integer = -1

    Const IOCTL_GET_HCD_DRIVERKEY_NAME As Integer = &H220424
    Const IOCTL_USB_GET_ROOT_HUB_NAME As Integer = &H220408
    Const IOCTL_USB_GET_NODE_INFORMATION As Integer = &H220408
    ' same as above... strange, eh?
    Const IOCTL_USB_GET_NODE_CONNECTION_INFORMATION_EX As Integer = &H220448
    Const IOCTL_USB_GET_DESCRIPTOR_FROM_NODE_CONNECTION As Integer = &H220410
    Const IOCTL_USB_GET_NODE_CONNECTION_NAME As Integer = &H220414
    Const IOCTL_USB_GET_NODE_CONNECTION_DRIVERKEY_NAME As Integer = &H220420

    Const USB_DEVICE_DESCRIPTOR_TYPE As Integer = &H1
    Const USB_STRING_DESCRIPTOR_TYPE As Integer = &H3

    Const BUFFER_SIZE As Integer = 2048
    Const MAXIMUM_USB_STRING_LENGTH As Integer = 255

    Const GUID_DEVINTERFACE_HUBCONTROLLER As String = "3abf6f2d-71c4-462a-8a92-1e6861e6af27"
    Const REGSTR_KEY_USB As String = "USB"
    Const DIGCF_PRESENT As Integer = &H2
    Const DIGCF_ALLCLASSES As Integer = &H4
    Const DIGCF_DEVICEINTERFACE As Integer = &H10
    Const SPDRP_DRIVER As Integer = &H9
    Const SPDRP_DEVICEDESC As Integer = &H0
    Const REG_SZ As Integer = 1

    ' ********************** Enumerations ************************

    'typedef enum _USB_HUB_NODE {
    '    UsbHub,
    '    UsbMIParent
    '} USB_HUB_NODE;
    Private Enum USB_HUB_NODE
        UsbHub
        UsbMIParent
    End Enum

    'typedef enum _USB_CONNECTION_STATUS {
    '    NoDeviceConnected,
    '    DeviceConnected,
    '    DeviceFailedEnumeration,
    '    DeviceGeneralFailure,
    '    DeviceCausedOvercurrent,
    '    DeviceNotEnoughPower,
    '    DeviceNotEnoughBandwidth,
    '    DeviceHubNestedTooDeeply,
    '    DeviceInLegacyHub
    '} USB_CONNECTION_STATUS, *PUSB_CONNECTION_STATUS;
    Private Enum USB_CONNECTION_STATUS
        NoDeviceConnected
        DeviceConnected
        DeviceFailedEnumeration
        DeviceGeneralFailure
        DeviceCausedOvercurrent
        DeviceNotEnoughPower
        DeviceNotEnoughBandwidth
        DeviceHubNestedTooDeeply
        DeviceInLegacyHub
    End Enum

    'typedef enum _USB_DEVICE_SPEED {
    '    UsbLowSpeed = 0,
    '    UsbFullSpeed,
    '    UsbHighSpeed
    '} USB_DEVICE_SPEED;
    Private Enum USB_DEVICE_SPEED As Byte
        UsbLowSpeed
        UsbFullSpeed
        UsbHighSpeed
    End Enum

    ' ********************** Stuctures ************************

    'typedef struct _SP_DEVINFO_DATA {
    '  DWORD cbSize;
    '  GUID ClassGuid;
    '  DWORD DevInst;
    '  ULONG_PTR Reserved;
    '} SP_DEVINFO_DATA,  *PSP_DEVINFO_DATA;
    <StructLayout(LayoutKind.Sequential)> _
    Private Structure SP_DEVINFO_DATA
        Public cbSize As Integer
        Public ClassGuid As Guid
        Public DevInst As Integer
        Public Reserved As IntPtr
    End Structure

    'typedef struct _SP_DEVICE_INTERFACE_DATA {
    '  DWORD cbSize;
    '  GUID InterfaceClassGuid;
    '  DWORD Flags;
    '  ULONG_PTR Reserved;
    '} SP_DEVICE_INTERFACE_DATA,  *PSP_DEVICE_INTERFACE_DATA;
    <StructLayout(LayoutKind.Sequential)> _
    Private Structure SP_DEVICE_INTERFACE_DATA
        Public cbSize As Integer
        Public InterfaceClassGuid As Guid
        Public Flags As Integer
        Public Reserved As IntPtr
    End Structure

    'typedef struct _SP_DEVICE_INTERFACE_DETAIL_DATA {
    '  DWORD cbSize;
    '  TCHAR DevicePath[ANYSIZE_ARRAY];
    '} SP_DEVICE_INTERFACE_DETAIL_DATA,  *PSP_DEVICE_INTERFACE_DETAIL_DATA;
    <StructLayout(LayoutKind.Sequential, CharSet:=CharSet.Auto)> _
    Private Structure SP_DEVICE_INTERFACE_DETAIL_DATA
        Public cbSize As Integer
        <MarshalAs(UnmanagedType.ByValTStr, SizeConst:=BUFFER_SIZE)> _
        Public DevicePath As String
    End Structure

    'typedef struct _USB_HCD_DRIVERKEY_NAME {
    '    ULONG ActualLength;
    '    WCHAR DriverKeyName[1];
    '} USB_HCD_DRIVERKEY_NAME, *PUSB_HCD_DRIVERKEY_NAME;
    <StructLayout(LayoutKind.Sequential, CharSet:=CharSet.Auto)> _
    Private Structure USB_HCD_DRIVERKEY_NAME
        Public ActualLength As Integer
        <MarshalAs(UnmanagedType.ByValTStr, SizeConst:=BUFFER_SIZE)> _
        Public DriverKeyName As String
    End Structure

    'typedef struct _USB_ROOT_HUB_NAME {
    '    ULONG  ActualLength;
    '    WCHAR  RootHubName[1];
    '} USB_ROOT_HUB_NAME, *PUSB_ROOT_HUB_NAME;
    <StructLayout(LayoutKind.Sequential, CharSet:=CharSet.Auto)> _
    Private Structure USB_ROOT_HUB_NAME
        Public ActualLength As Integer
        <MarshalAs(UnmanagedType.ByValTStr, SizeConst:=BUFFER_SIZE)> _
        Public RootHubName As String
    End Structure

    'typedef struct _USB_HUB_DESCRIPTOR {
    '    UCHAR  bDescriptorLength;
    '    UCHAR  bDescriptorType;
    '    UCHAR  bNumberOfPorts;
    '    USHORT  wHubCharacteristics;
    '    UCHAR  bPowerOnToPowerGood;
    '    UCHAR  bHubControlCurrent;
    '    UCHAR  bRemoveAndPowerMask[64];
    '} USB_HUB_DESCRIPTOR, *PUSB_HUB_DESCRIPTOR;
    <StructLayout(LayoutKind.Sequential, Pack:=1)> _
    Private Structure USB_HUB_DESCRIPTOR
        Public bDescriptorLength As Byte
        Public bDescriptorType As Byte
        Public bNumberOfPorts As Byte
        Public wHubCharacteristics As Short
        Public bPowerOnToPowerGood As Byte
        Public bHubControlCurrent As Byte
        <MarshalAs(UnmanagedType.ByValArray, SizeConst:=64)> _
        Public bRemoveAndPowerMask As Byte()
    End Structure

    'typedef struct _USB_HUB_INFORMATION {
    '    USB_HUB_DESCRIPTOR HubDescriptor;
    '    BOOLEAN HubIsBusPowered;
    '} USB_HUB_INFORMATION, *PUSB_HUB_INFORMATION;
    <StructLayout(LayoutKind.Sequential)> _
    Private Structure USB_HUB_INFORMATION
        Public HubDescriptor As USB_HUB_DESCRIPTOR
        Public HubIsBusPowered As Byte
    End Structure

    'typedef struct _USB_NODE_INFORMATION {
    '    USB_HUB_NODE  NodeType;
    '    union {
    '        USB_HUB_INFORMATION  HubInformation;
    '        USB_MI_PARENT_INFORMATION  MiParentInformation;
    '    } u;
    '} USB_NODE_INFORMATION, *PUSB_NODE_INFORMATION;
    <StructLayout(LayoutKind.Sequential)> _
    Private Structure USB_NODE_INFORMATION
        Public NodeType As Integer
        Public HubInformation As USB_HUB_INFORMATION
        ' Yeah, I'm assuming we'll just use the first form
    End Structure

    'typedef struct _USB_NODE_CONNECTION_INFORMATION_EX {
    '    ULONG  ConnectionIndex;
    '    USB_DEVICE_DESCRIPTOR  DeviceDescriptor;
    '    UCHAR  CurrentConfigurationValue;
    '    UCHAR  Speed;
    '    BOOLEAN  DeviceIsHub;
    '    USHORT  DeviceAddress;
    '    ULONG  NumberOfOpenPipes;
    '    USB_CONNECTION_STATUS  ConnectionStatus;
    '    USB_PIPE_INFO  PipeList[0];
    '} USB_NODE_CONNECTION_INFORMATION_EX, *PUSB_NODE_CONNECTION_INFORMATION_EX;
    <StructLayout(LayoutKind.Sequential, Pack:=1)> _
    Private Structure USB_NODE_CONNECTION_INFORMATION_EX
        Public ConnectionIndex As Integer
        Public DeviceDescriptor As USB_DEVICE_DESCRIPTOR
        Public CurrentConfigurationValue As Byte
        Public Speed As Byte
        Public DeviceIsHub As Byte
        Public DeviceAddress As Short
        Public NumberOfOpenPipes As Integer
        Public ConnectionStatus As Integer
        'public IntPtr PipeList;
    End Structure

    'typedef struct _USB_DEVICE_DESCRIPTOR {
    '    UCHAR  bLength;
    '    UCHAR  bDescriptorType;
    '    USHORT  bcdUSB;
    '    UCHAR  bDeviceClass;
    '    UCHAR  bDeviceSubClass;
    '    UCHAR  bDeviceProtocol;
    '    UCHAR  bMaxPacketSize0;
    '    USHORT  idVendor;
    '    USHORT  idProduct;
    '    USHORT  bcdDevice;
    '    UCHAR  iManufacturer;
    '    UCHAR  iProduct;
    '    UCHAR  iSerialNumber;
    '    UCHAR  bNumConfigurations;
    '} USB_DEVICE_DESCRIPTOR, *PUSB_DEVICE_DESCRIPTOR ;
    <StructLayout(LayoutKind.Sequential, Pack:=1)> _
    Friend Structure USB_DEVICE_DESCRIPTOR
        Public bLength As Byte
        Public bDescriptorType As Byte
        Public bcdUSB As Short
        Public bDeviceClass As Byte
        Public bDeviceSubClass As Byte
        Public bDeviceProtocol As Byte
        Public bMaxPacketSize0 As Byte
        Public idVendor As Short
        Public idProduct As Short
        Public bcdDevice As Short
        Public iManufacturer As Byte
        Public iProduct As Byte
        Public iSerialNumber As Byte
        Public bNumConfigurations As Byte
    End Structure

    'typedef struct _USB_STRING_DESCRIPTOR {
    '    UCHAR bLength;
    '    UCHAR bDescriptorType;
    '    WCHAR bString[1];
    '} USB_STRING_DESCRIPTOR, *PUSB_STRING_DESCRIPTOR;
    <StructLayout(LayoutKind.Sequential, CharSet:=CharSet.Auto)> _
    Private Structure USB_STRING_DESCRIPTOR
        Public bLength As Byte
        Public bDescriptorType As Byte
        <MarshalAs(UnmanagedType.ByValTStr, SizeConst:=MAXIMUM_USB_STRING_LENGTH)> _
        Public bString As String
    End Structure

    'typedef struct _USB_DESCRIPTOR_REQUEST {
    '  ULONG ConnectionIndex;
    '  struct {
    '    UCHAR  bmRequest;
    '    UCHAR  bRequest;
    '    USHORT  wValue;
    '    USHORT  wIndex;
    '    USHORT  wLength;
    '  } SetupPacket;
    '  UCHAR  Data[0];
    '} USB_DESCRIPTOR_REQUEST, *PUSB_DESCRIPTOR_REQUEST
    <StructLayout(LayoutKind.Sequential)> _
    Private Structure USB_SETUP_PACKET
        Public bmRequest As Byte
        Public bRequest As Byte
        Public wValue As Short
        Public wIndex As Short
        Public wLength As Short
    End Structure
    <StructLayout(LayoutKind.Sequential)> _
    Private Structure USB_DESCRIPTOR_REQUEST
        Public ConnectionIndex As Integer
        Public SetupPacket As USB_SETUP_PACKET
        'public byte[] Data;
    End Structure

    'typedef struct _USB_NODE_CONNECTION_NAME {
    '    ULONG  ConnectionIndex;
    '    ULONG  ActualLength;
    '    WCHAR  NodeName[1];
    '} USB_NODE_CONNECTION_NAME, *PUSB_NODE_CONNECTION_NAME;
    <StructLayout(LayoutKind.Sequential, CharSet:=CharSet.Auto)> _
    Private Structure USB_NODE_CONNECTION_NAME
        Public ConnectionIndex As Integer
        Public ActualLength As Integer
        <MarshalAs(UnmanagedType.ByValTStr, SizeConst:=BUFFER_SIZE)> _
        Public NodeName As String
    End Structure

    'typedef struct _USB_NODE_CONNECTION_DRIVERKEY_NAME {
    '    ULONG  ConnectionIndex;
    '    ULONG  ActualLength;
    '    WCHAR  DriverKeyName[1];
    '} USB_NODE_CONNECTION_DRIVERKEY_NAME, *PUSB_NODE_CONNECTION_DRIVERKEY_NAME;
    <StructLayout(LayoutKind.Sequential, CharSet:=CharSet.Auto)> _
    Private Structure USB_NODE_CONNECTION_DRIVERKEY_NAME
        ' Yes, this is the same as the structure above...
        Public ConnectionIndex As Integer
        Public ActualLength As Integer
        <MarshalAs(UnmanagedType.ByValTStr, SizeConst:=BUFFER_SIZE)> _
        Public DriverKeyName As String
    End Structure

    ' ********************** API Definitions ************************

    'HDEVINFO SetupDiGetClassDevs(
    '  const GUID* ClassGuid,
    '  PCTSTR Enumerator,
    '  HWND hwndParent,
    '  DWORD Flags
    ');
    ' 1st form using a ClassGUID
    <DllImport("setupapi.dll", CharSet:=CharSet.Auto)> _
    Private Shared Function SetupDiGetClassDevs(ByRef ClassGuid As Guid, ByVal Enumerator As Integer, ByVal hwndParent As IntPtr, ByVal Flags As Integer) As IntPtr
    End Function
    ' 2nd form uses an Enumerator
    <DllImport("setupapi.dll", CharSet:=CharSet.Auto)> _
    Private Shared Function SetupDiGetClassDevs(ByVal ClassGuid As Integer, ByVal Enumerator As String, ByVal hwndParent As IntPtr, ByVal Flags As Integer) As IntPtr
    End Function

    'BOOL SetupDiEnumDeviceInterfaces(
    '  HDEVINFO DeviceInfoSet,
    '  PSP_DEVINFO_DATA DeviceInfoData,
    '  const GUID* InterfaceClassGuid,
    '  DWORD MemberIndex,
    '  PSP_DEVICE_INTERFACE_DATA DeviceInterfaceData
    ');
    <DllImport("setupapi.dll", SetLastError:=True, CharSet:=CharSet.Auto)> _
    Private Shared Function SetupDiEnumDeviceInterfaces(ByVal DeviceInfoSet As IntPtr, ByVal DeviceInfoData As IntPtr, ByRef InterfaceClassGuid As Guid, ByVal MemberIndex As Integer, ByRef DeviceInterfaceData As SP_DEVICE_INTERFACE_DATA) As Boolean
    End Function

    'BOOL SetupDiGetDeviceInterfaceDetail(
    '  HDEVINFO DeviceInfoSet,
    '  PSP_DEVICE_INTERFACE_DATA DeviceInterfaceData,
    '  PSP_DEVICE_INTERFACE_DETAIL_DATA DeviceInterfaceDetailData,
    '  DWORD DeviceInterfaceDetailDataSize,
    '  PDWORD RequiredSize,
    '  PSP_DEVINFO_DATA DeviceInfoData
    ');
    <DllImport("setupapi.dll", SetLastError:=True, CharSet:=CharSet.Auto)> _
    Private Shared Function SetupDiGetDeviceInterfaceDetail(ByVal DeviceInfoSet As IntPtr, ByRef DeviceInterfaceData As SP_DEVICE_INTERFACE_DATA, ByRef DeviceInterfaceDetailData As SP_DEVICE_INTERFACE_DETAIL_DATA, ByVal DeviceInterfaceDetailDataSize As Integer, ByRef RequiredSize As Integer, ByRef DeviceInfoData As SP_DEVINFO_DATA) As Boolean
    End Function

    'BOOL SetupDiGetDeviceRegistryProperty(
    '  HDEVINFO DeviceInfoSet,
    '  PSP_DEVINFO_DATA DeviceInfoData,
    '  DWORD Property,
    '  PDWORD PropertyRegDataType,
    '  PBYTE PropertyBuffer,
    '  DWORD PropertyBufferSize,
    '  PDWORD RequiredSize
    ');
    <DllImport("setupapi.dll", SetLastError:=True, CharSet:=CharSet.Auto)> _
    Private Shared Function SetupDiGetDeviceRegistryProperty(ByVal DeviceInfoSet As IntPtr, ByRef DeviceInfoData As SP_DEVINFO_DATA, ByVal iProperty As Integer, ByRef PropertyRegDataType As Integer, ByVal PropertyBuffer As IntPtr, ByVal PropertyBufferSize As Integer, _
 ByRef RequiredSize As Integer) As Boolean
    End Function

    'BOOL SetupDiEnumDeviceInfo(
    '  HDEVINFO DeviceInfoSet,
    '  DWORD MemberIndex,
    '  PSP_DEVINFO_DATA DeviceInfoData
    ');
    <DllImport("setupapi.dll", SetLastError:=True, CharSet:=CharSet.Auto)> _
    Private Shared Function SetupDiEnumDeviceInfo(ByVal DeviceInfoSet As IntPtr, ByVal MemberIndex As Integer, ByRef DeviceInfoData As SP_DEVINFO_DATA) As Boolean
    End Function

    'BOOL SetupDiDestroyDeviceInfoList(
    '  HDEVINFO DeviceInfoSet
    ');
    <DllImport("setupapi.dll", SetLastError:=True)> _
    Private Shared Function SetupDiDestroyDeviceInfoList(ByVal DeviceInfoSet As IntPtr) As Boolean
    End Function

    'WINSETUPAPI BOOL WINAPI  SetupDiGetDeviceInstanceId(
    '    IN HDEVINFO  DeviceInfoSet,
    '    IN PSP_DEVINFO_DATA  DeviceInfoData,
    '    OUT PTSTR  DeviceInstanceId,
    '    IN DWORD  DeviceInstanceIdSize,
    '    OUT PDWORD  RequiredSize  OPTIONAL
    ');
    <DllImport("setupapi.dll", SetLastError:=True, CharSet:=CharSet.Auto)> _
    Private Shared Function SetupDiGetDeviceInstanceId(ByVal DeviceInfoSet As IntPtr, ByRef DeviceInfoData As SP_DEVINFO_DATA, ByVal DeviceInstanceId As StringBuilder, ByVal DeviceInstanceIdSize As Integer, ByRef RequiredSize As Integer) As Boolean
    End Function

    'BOOL DeviceIoControl(
    '  HANDLE hDevice,
    '  DWORD dwIoControlCode,
    '  LPVOID lpInBuffer,
    '  DWORD nInBufferSize,
    '  LPVOID lpOutBuffer,
    '  DWORD nOutBufferSize,
    '  LPDWORD lpBytesReturned,
    '  LPOVERLAPPED lpOverlapped
    ');
    <DllImport("kernel32.dll", SetLastError:=True, CharSet:=CharSet.Auto)> _
    Private Shared Function DeviceIoControl(ByVal hDevice As IntPtr, ByVal dwIoControlCode As Integer, ByVal lpInBuffer As IntPtr, ByVal nInBufferSize As Integer, ByVal lpOutBuffer As IntPtr, ByVal nOutBufferSize As Integer, _
 ByRef lpBytesReturned As Integer, ByVal lpOverlapped As IntPtr) As Boolean
    End Function

    'HANDLE CreateFile(
    '  LPCTSTR lpFileName,
    '  DWORD dwDesiredAccess,
    '  DWORD dwShareMode,
    '  LPSECURITY_ATTRIBUTES lpSecurityAttributes,
    '  DWORD dwCreationDisposition,
    '  DWORD dwFlagsAndAttributes,
    '  HANDLE hTemplateFile
    ');
    <DllImport("kernel32.dll", SetLastError:=True, CharSet:=CharSet.Auto)> _
    Private Shared Function CreateFile(ByVal lpFileName As String, ByVal dwDesiredAccess As Integer, ByVal dwShareMode As Integer, ByVal lpSecurityAttributes As IntPtr, ByVal dwCreationDisposition As Integer, ByVal dwFlagsAndAttributes As Integer, _
 ByVal hTemplateFile As IntPtr) As IntPtr
    End Function

    'BOOL CloseHandle(
    '  HANDLE hObject
    ');
    <DllImport("kernel32.dll", SetLastError:=True, CharSet:=CharSet.Auto)> _
    Private Shared Function CloseHandle(ByVal hObject As IntPtr) As Boolean
    End Function
#End Region

    Public Shared Function GetConnectedDevices() As List(Of USBDevice)
        Dim DevList As New List(Of USBDevice)()

        For Each Controller As USBController In GetHostControllers()
            ListHub(Controller.GetRootHub(), DevList)
        Next
        Return DevList
    End Function

    Private Shared Sub ListHub(ByVal Hub As USBHub, ByVal DevList As List(Of USBDevice))
        For Each Port As USBPort In Hub.GetPorts()
            If Port.IsHub Then
                ' recursive
                ListHub(Port.GetHub(), DevList)
            Else
                If Port.IsDeviceConnected Then
                    DevList.Add(Port.GetDevice())
                End If
            End If
        Next
    End Sub

    Private Shared Sub SearchHubInstanceID(ByVal Hub As USBHub, ByRef FoundDevice As USBDevice, ByVal InstanceID As String)
        For Each Port As USBPort In Hub.GetPorts()
            If Port.IsHub Then
                ' recursive
                SearchHubInstanceID(Port.GetHub(), FoundDevice, InstanceID)
            Else
                If Port.IsDeviceConnected Then
                    Dim Device As USBDevice = Port.GetDevice()
                    If Device.InstanceID = InstanceID Then
                        FoundDevice = Device
                        Exit For
                    End If
                End If
            End If
        Next
    End Sub

    Public Shared Function GetHostControllers() As System.Collections.ObjectModel.ReadOnlyCollection(Of USBController)
        Dim HostList As New List(Of USBController)()
        Dim HostGUID As New Guid(GUID_DEVINTERFACE_HUBCONTROLLER)

        ' We start at the "root" of the device tree and look for all
        ' devices that match the interface GUID of a Hub Controller
        Dim h As IntPtr = SetupDiGetClassDevs(HostGUID, 0, IntPtr.Zero, DIGCF_PRESENT Or DIGCF_DEVICEINTERFACE)
        If h.ToInt32() <> INVALID_HANDLE_VALUE Then
            Dim ptrBuf As IntPtr = Marshal.AllocHGlobal(BUFFER_SIZE)
            Dim Success As Boolean
            Dim i As Integer = 0
            Do
                Dim host As New USBController()
                host.ControllerIndex = i

                ' create a Device Interface Data structure
                Dim dia As New SP_DEVICE_INTERFACE_DATA()
                dia.cbSize = Marshal.SizeOf(dia)

                ' start the enumeration 
                Success = SetupDiEnumDeviceInterfaces(h, IntPtr.Zero, HostGUID, i, dia)
                If Success Then
                    ' build a DevInfo Data structure
                    Dim da As New SP_DEVINFO_DATA()
                    da.cbSize = Marshal.SizeOf(da)

                    ' build a Device Interface Detail Data structure
                    Dim didd As New SP_DEVICE_INTERFACE_DETAIL_DATA()
                    didd.cbSize = If((IntPtr.Size = 4), (4 + Marshal.SystemDefaultCharSize), 8)
                    ' trust me :)
                    ' now we can get some more detailed information
                    Dim nRequiredSize As Integer = 0
                    Dim nBytes As Integer = BUFFER_SIZE
                    If SetupDiGetDeviceInterfaceDetail(h, dia, didd, nBytes, nRequiredSize, da) Then
                        host.ControllerDevicePath = didd.DevicePath

                        ' get the Device Description and DriverKeyName
                        Dim RequiredSize As Integer = 0
                        Dim RegType As Integer = REG_SZ

                        If SetupDiGetDeviceRegistryProperty(h, da, SPDRP_DEVICEDESC, RegType, ptrBuf, BUFFER_SIZE, _
                         RequiredSize) Then
                            host.ControllerDeviceDesc = Marshal.PtrToStringAuto(ptrBuf)
                        End If
                        If SetupDiGetDeviceRegistryProperty(h, da, SPDRP_DRIVER, RegType, ptrBuf, BUFFER_SIZE, _
                         RequiredSize) Then
                            host.ControllerDriverKeyName = Marshal.PtrToStringAuto(ptrBuf)
                        End If
                    End If
                    HostList.Add(host)
                End If
                i += 1
            Loop While Success

            Marshal.FreeHGlobal(ptrBuf)
            SetupDiDestroyDeviceInfoList(h)
        End If

        ' convert it into a Collection
        Return New System.Collections.ObjectModel.ReadOnlyCollection(Of USBController)(HostList)
    End Function

    'The USB Host Controller Class
    Public Class USBController
        Friend ControllerIndex As Integer
        Friend ControllerDriverKeyName As String, ControllerDevicePath As String, ControllerDeviceDesc As String

        ' A simple default constructor
        Public Sub New()
            ControllerIndex = 0
            ControllerDevicePath = ""
            ControllerDeviceDesc = ""
            ControllerDriverKeyName = ""
        End Sub

        ' Return the index of the instance
        Public ReadOnly Property Index() As Integer
            Get
                Return ControllerIndex
            End Get
        End Property

        ' Return the Device Path, such as "\\?\pci#ven_10de&dev_005a&subsys_815a1043&rev_a2#3&267a616a&0&58#{3abf6f2d-71c4-462a-8a92-1e6861e6af27}"
        Public ReadOnly Property DevicePath() As String
            Get
                Return ControllerDevicePath
            End Get
        End Property

        ' The DriverKeyName may be useful as a search key
        Public ReadOnly Property DriverKeyName() As String
            Get
                Return ControllerDriverKeyName
            End Get
        End Property

        ' Return the Friendly Name, such as "VIA USB Enhanced Host Controller"
        Public ReadOnly Property Name() As String
            Get
                Return ControllerDeviceDesc
            End Get
        End Property



        ' Return Root Hub for this Controller
        Public Function GetRootHub() As USBHub
            Dim h As IntPtr, h2 As IntPtr
            Dim Root As New USBHub()
            Root.HubIsRootHub = True
            Root.HubDeviceDesc = "Root Hub"

            ' Open a handle to the Host Controller
            h = CreateFile(ControllerDevicePath, GENERIC_WRITE, FILE_SHARE_WRITE, IntPtr.Zero, OPEN_EXISTING, 0, _
             IntPtr.Zero)
            If h.ToInt32() <> INVALID_HANDLE_VALUE Then
                Dim nBytesReturned As Integer
                Dim HubName As New USB_ROOT_HUB_NAME()
                Dim nBytes As Integer = Marshal.SizeOf(HubName)
                Dim ptrHubName As IntPtr = Marshal.AllocHGlobal(nBytes)

                ' get the Hub Name
                If DeviceIoControl(h, IOCTL_USB_GET_ROOT_HUB_NAME, ptrHubName, nBytes, ptrHubName, nBytes, _
                 nBytesReturned, IntPtr.Zero) Then
                    HubName = DirectCast(Marshal.PtrToStructure(ptrHubName, GetType(USB_ROOT_HUB_NAME)), USB_ROOT_HUB_NAME)
                    Root.HubDevicePath = "\\.\" + HubName.RootHubName
                End If

                ' TODO: Get DriverKeyName for Root Hub

                ' Now let's open the Hub (based upon the HubName we got above)
                h2 = CreateFile(Root.HubDevicePath, GENERIC_WRITE, FILE_SHARE_WRITE, IntPtr.Zero, OPEN_EXISTING, 0, _
                 IntPtr.Zero)
                If h2.ToInt32() <> INVALID_HANDLE_VALUE Then
                    Dim NodeInfo As New USB_NODE_INFORMATION()
                    NodeInfo.NodeType = CInt(USB_HUB_NODE.UsbHub)
                    nBytes = Marshal.SizeOf(NodeInfo)
                    Dim ptrNodeInfo As IntPtr = Marshal.AllocHGlobal(nBytes)
                    Marshal.StructureToPtr(NodeInfo, ptrNodeInfo, True)

                    ' get the Hub Information
                    If DeviceIoControl(h2, IOCTL_USB_GET_NODE_INFORMATION, ptrNodeInfo, nBytes, ptrNodeInfo, nBytes, _
                     nBytesReturned, IntPtr.Zero) Then
                        NodeInfo = DirectCast(Marshal.PtrToStructure(ptrNodeInfo, GetType(USB_NODE_INFORMATION)), USB_NODE_INFORMATION)
                        Root.HubIsBusPowered = Convert.ToBoolean(NodeInfo.HubInformation.HubIsBusPowered)
                        Root.HubPortCount = NodeInfo.HubInformation.HubDescriptor.bNumberOfPorts
                    End If
                    Marshal.FreeHGlobal(ptrNodeInfo)
                    CloseHandle(h2)
                End If

                Marshal.FreeHGlobal(ptrHubName)
                CloseHandle(h)
            End If
            Return Root
        End Function
    End Class

    ' The USB Device class
    Public Class USBDevice
        Friend DevicePortNumber As Integer
        Friend DeviceDriverKey As String, DeviceHubDevicePath As String, DeviceInstanceID As String, DeviceName As String
        Friend DeviceManufacturer As String, DeviceProduct As String, DeviceSerialNumber As String
        Friend DeviceDescriptor As USB_DEVICE_DESCRIPTOR

        ' a simple default constructor
        Public Sub New()
            DevicePortNumber = 0
            DeviceHubDevicePath = ""
            DeviceDriverKey = ""
            DeviceManufacturer = ""
            DeviceProduct = "Unknown USB Device"
            DeviceSerialNumber = ""
            DeviceName = ""
            DeviceInstanceID = ""
        End Sub

        ' return Port Index of the Hub
        Public ReadOnly Property PortNumber() As Integer
            Get
                Return DevicePortNumber
            End Get
        End Property

        ' return the Device Path of the Hub (the parent device)
        Public ReadOnly Property HubDevicePath() As String
            Get
                Return DeviceHubDevicePath
            End Get
        End Property

        ' useful as a search key
        Public ReadOnly Property DriverKey() As String
            Get
                Return DeviceDriverKey
            End Get
        End Property

        ' the device path of this device
        Public ReadOnly Property InstanceID() As String
            Get
                Return DeviceInstanceID
            End Get
        End Property

        ' the friendly name
        Public ReadOnly Property Name() As String
            Get
                Return DeviceName
            End Get
        End Property

        Public ReadOnly Property Manufacturer() As String
            Get
                Return DeviceManufacturer
            End Get
        End Property

        Public ReadOnly Property Product() As String
            Get
                Return DeviceProduct
            End Get
        End Property

        Public ReadOnly Property SerialNumber() As String
            Get
                Return DeviceSerialNumber
            End Get
        End Property
    End Class

    ' The Hub class
    Public Class USBHub
        Friend HubPortCount As Integer
        Friend HubDriverKey As String, HubDevicePath As String, HubDeviceDesc As String
        Friend HubManufacturer As String, HubProduct As String, HubSerialNumber As String, HubInstanceID As String
        Friend HubIsBusPowered As Boolean, HubIsRootHub As Boolean

        ' a simple default constructor
        Public Sub New()
            HubPortCount = 0
            HubDevicePath = ""
            HubDeviceDesc = ""
            HubDriverKey = ""
            HubIsBusPowered = False
            HubIsRootHub = False
            HubManufacturer = ""
            HubProduct = ""
            HubSerialNumber = ""
            HubInstanceID = ""
        End Sub

        ' return Port Count
        Public ReadOnly Property PortCount() As Integer
            Get
                Return HubPortCount
            End Get
        End Property

        ' return the Device Path, such as "\\?\pci#ven_10de&dev_005a&subsys_815a1043&rev_a2#3&267a616a&0&58#{3abf6f2d-71c4-462a-8a92-1e6861e6af27}"
        Public ReadOnly Property DevicePath() As String
            Get
                Return HubDevicePath
            End Get
        End Property

        ' The DriverKey may be useful as a search key
        Public ReadOnly Property DriverKey() As String
            Get
                Return HubDriverKey
            End Get
        End Property

        ' return the Friendly Name, such as "VIA USB Enhanced Host Controller"
        Public ReadOnly Property Name() As String
            Get
                Return HubDeviceDesc
            End Get
        End Property

        ' the device path of this device
        Public ReadOnly Property InstanceID() As String
            Get
                Return HubInstanceID
            End Get
        End Property

        ' is is this a self-powered hub?
        Public ReadOnly Property IsBusPowered() As Boolean
            Get
                Return HubIsBusPowered
            End Get
        End Property

        ' is this a root hub?
        Public ReadOnly Property IsRootHub() As Boolean
            Get
                Return HubIsRootHub
            End Get
        End Property

        Public ReadOnly Property Manufacturer() As String
            Get
                Return HubManufacturer
            End Get
        End Property

        Public ReadOnly Property Product() As String
            Get
                Return HubProduct
            End Get
        End Property

        Public ReadOnly Property SerialNumber() As String
            Get
                Return HubSerialNumber
            End Get
        End Property

        ' return a list of the down stream ports
        Public Function GetPorts() As System.Collections.ObjectModel.ReadOnlyCollection(Of USBPort)
            Dim PortList As New List(Of USBPort)()

            ' Open a handle to the Hub device
            Dim h As IntPtr = CreateFile(HubDevicePath, GENERIC_WRITE, FILE_SHARE_WRITE, IntPtr.Zero, OPEN_EXISTING, 0, _
             IntPtr.Zero)
            If h.ToInt32() <> INVALID_HANDLE_VALUE Then
                Dim nBytes As Integer = Marshal.SizeOf(GetType(USB_NODE_CONNECTION_INFORMATION_EX))
                Dim ptrNodeConnection As IntPtr = Marshal.AllocHGlobal(nBytes)

                ' loop thru all of the ports on the hub
                ' BTW: Ports are numbered starting at 1
                For i As Integer = 1 To HubPortCount
                    Dim nBytesReturned As Integer
                    Dim NodeConnection As New USB_NODE_CONNECTION_INFORMATION_EX()
                    NodeConnection.ConnectionIndex = i
                    Marshal.StructureToPtr(NodeConnection, ptrNodeConnection, True)

                    If DeviceIoControl(h, IOCTL_USB_GET_NODE_CONNECTION_INFORMATION_EX, ptrNodeConnection, nBytes, ptrNodeConnection, nBytes, _
                     nBytesReturned, IntPtr.Zero) Then
                        NodeConnection = DirectCast(Marshal.PtrToStructure(ptrNodeConnection, GetType(USB_NODE_CONNECTION_INFORMATION_EX)), USB_NODE_CONNECTION_INFORMATION_EX)

                        ' load up the USBPort class
                        Dim port As New USBPort()
                        port.PortPortNumber = i
                        port.PortHubDevicePath = HubDevicePath
                        Dim Status As USB_CONNECTION_STATUS = DirectCast(NodeConnection.ConnectionStatus, USB_CONNECTION_STATUS)
                        port.PortStatus = Status.ToString()
                        Dim Speed As USB_DEVICE_SPEED = DirectCast(NodeConnection.Speed, USB_DEVICE_SPEED)
                        port.PortSpeed = Speed.ToString()
                        port.PortIsDeviceConnected = (NodeConnection.ConnectionStatus = CInt(USB_CONNECTION_STATUS.DeviceConnected))
                        port.PortIsHub = Convert.ToBoolean(NodeConnection.DeviceIsHub)
                        port.PortDeviceDescriptor = NodeConnection.DeviceDescriptor

                        ' add it to the list
                        PortList.Add(port)
                    End If
                Next
                Marshal.FreeHGlobal(ptrNodeConnection)
                CloseHandle(h)
            End If
            ' convert it into a Collection
            Return New System.Collections.ObjectModel.ReadOnlyCollection(Of USBPort)(PortList)
        End Function
    End Class

    ' The Port Class
    Public Class USBPort
        Friend PortPortNumber As Integer
        Friend PortStatus As String, PortHubDevicePath As String, PortSpeed As String
        Friend PortIsHub As Boolean, PortIsDeviceConnected As Boolean
        Friend PortDeviceDescriptor As USB_DEVICE_DESCRIPTOR

        ' a simple default constructor
        Public Sub New()
            PortPortNumber = 0
            PortStatus = ""
            PortHubDevicePath = ""
            PortSpeed = ""
            PortIsHub = False
            PortIsDeviceConnected = False
        End Sub

        ' return Port Index of the Hub
        Public ReadOnly Property PortNumber() As Integer
            Get
                Return PortPortNumber
            End Get
        End Property

        ' return the Device Path of the Hub
        Public ReadOnly Property HubDevicePath() As String
            Get
                Return PortHubDevicePath
            End Get
        End Property

        ' the status (see USB_CONNECTION_STATUS above)
        Public ReadOnly Property Status() As String
            Get
                Return PortStatus
            End Get
        End Property

        ' the speed of the connection (see USB_DEVICE_SPEED above)
        Public ReadOnly Property Speed() As String
            Get
                Return PortSpeed
            End Get
        End Property

        ' is this a downstream external hub?
        Public ReadOnly Property IsHub() As Boolean
            Get
                Return PortIsHub
            End Get
        End Property

        ' is anybody home?
        Public ReadOnly Property IsDeviceConnected() As Boolean
            Get
                Return PortIsDeviceConnected
            End Get
        End Property

        ' return a down stream external hub
        Public Function GetDevice() As USBDevice
            If Not PortIsDeviceConnected Then
                Return Nothing
            End If
            Dim Device As New USBDevice()

            ' Copy over some values from the Port class
            ' Ya know, I've given some thought about making Device a derived class...
            Device.DevicePortNumber = PortPortNumber
            Device.DeviceHubDevicePath = PortHubDevicePath
            Device.DeviceDescriptor = PortDeviceDescriptor

            ' Open a handle to the Hub device
            Dim h As IntPtr = CreateFile(PortHubDevicePath, GENERIC_WRITE, FILE_SHARE_WRITE, IntPtr.Zero, OPEN_EXISTING, 0, _
             IntPtr.Zero)
            If h.ToInt32() <> INVALID_HANDLE_VALUE Then
                Dim nBytesReturned As Integer
                Dim nBytes As Integer = BUFFER_SIZE
                ' We use this to zero fill a buffer
                Dim NullString As New String(ChrW(0), BUFFER_SIZE / Marshal.SystemDefaultCharSize)

                ' The iManufacturer, iProduct and iSerialNumber entries in the
                ' Device Descriptor are really just indexes.  So, we have to 
                ' request a String Descriptor to get the values for those strings.

                If PortDeviceDescriptor.iManufacturer > 0 Then
                    ' build a request for string descriptor
                    Dim Request As New USB_DESCRIPTOR_REQUEST()
                    Request.ConnectionIndex = PortPortNumber
                    Request.SetupPacket.wValue = CShort((USB_STRING_DESCRIPTOR_TYPE << 8) + PortDeviceDescriptor.iManufacturer)
                    Request.SetupPacket.wLength = CShort(nBytes - Marshal.SizeOf(Request))
                    Request.SetupPacket.wIndex = &H409
                    ' Language Code
                    ' Geez, I wish C# had a Marshal.MemSet() method
                    Dim ptrRequest As IntPtr = Marshal.StringToHGlobalAuto(NullString)
                    Marshal.StructureToPtr(Request, ptrRequest, True)

                    ' Use an IOCTL call to request the String Descriptor
                    If DeviceIoControl(h, IOCTL_USB_GET_DESCRIPTOR_FROM_NODE_CONNECTION, ptrRequest, nBytes, ptrRequest, nBytes, _
                     nBytesReturned, IntPtr.Zero) Then
                        ' The location of the string descriptor is immediately after
                        ' the Request structure.  Because this location is not "covered"
                        ' by the structure allocation, we're forced to zero out this
                        ' chunk of memory by using the StringToHGlobalAuto() hack above
                        Dim ptrStringDesc As New IntPtr(ptrRequest.ToInt32() + Marshal.SizeOf(Request))
                        Dim StringDesc As USB_STRING_DESCRIPTOR = DirectCast(Marshal.PtrToStructure(ptrStringDesc, GetType(USB_STRING_DESCRIPTOR)), USB_STRING_DESCRIPTOR)
                        Device.DeviceManufacturer = StringDesc.bString
                    End If
                    Marshal.FreeHGlobal(ptrRequest)
                End If
                If PortDeviceDescriptor.iProduct > 0 Then
                    ' build a request for string descriptor
                    Dim Request As New USB_DESCRIPTOR_REQUEST()
                    Request.ConnectionIndex = PortPortNumber
                    Request.SetupPacket.wValue = CShort((USB_STRING_DESCRIPTOR_TYPE << 8) + PortDeviceDescriptor.iProduct)
                    Request.SetupPacket.wLength = CShort(nBytes - Marshal.SizeOf(Request))
                    Request.SetupPacket.wIndex = &H409
                    ' Language Code
                    ' Geez, I wish C# had a Marshal.MemSet() method
                    Dim ptrRequest As IntPtr = Marshal.StringToHGlobalAuto(NullString)
                    Marshal.StructureToPtr(Request, ptrRequest, True)

                    ' Use an IOCTL call to request the String Descriptor
                    If DeviceIoControl(h, IOCTL_USB_GET_DESCRIPTOR_FROM_NODE_CONNECTION, ptrRequest, nBytes, ptrRequest, nBytes, _
                     nBytesReturned, IntPtr.Zero) Then
                        ' the location of the string descriptor is immediately after the Request structure
                        Dim ptrStringDesc As New IntPtr(ptrRequest.ToInt32() + Marshal.SizeOf(Request))
                        Dim StringDesc As USB_STRING_DESCRIPTOR = DirectCast(Marshal.PtrToStructure(ptrStringDesc, GetType(USB_STRING_DESCRIPTOR)), USB_STRING_DESCRIPTOR)
                        Device.DeviceProduct = StringDesc.bString
                    End If
                    Marshal.FreeHGlobal(ptrRequest)
                End If
                If PortDeviceDescriptor.iSerialNumber > 0 Then
                    ' build a request for string descriptor
                    Dim Request As New USB_DESCRIPTOR_REQUEST()
                    Request.ConnectionIndex = PortPortNumber
                    Request.SetupPacket.wValue = CShort((USB_STRING_DESCRIPTOR_TYPE << 8) + PortDeviceDescriptor.iSerialNumber)
                    Request.SetupPacket.wLength = CShort(nBytes - Marshal.SizeOf(Request))
                    Request.SetupPacket.wIndex = &H409
                    ' Language Code
                    ' Geez, I wish C# had a Marshal.MemSet() method
                    Dim ptrRequest As IntPtr = Marshal.StringToHGlobalAuto(NullString)
                    Marshal.StructureToPtr(Request, ptrRequest, True)

                    ' Use an IOCTL call to request the String Descriptor
                    If DeviceIoControl(h, IOCTL_USB_GET_DESCRIPTOR_FROM_NODE_CONNECTION, ptrRequest, nBytes, ptrRequest, nBytes, _
                     nBytesReturned, IntPtr.Zero) Then
                        ' the location of the string descriptor is immediately after the Request structure
                        Dim ptrStringDesc As New IntPtr(ptrRequest.ToInt32() + Marshal.SizeOf(Request))
                        Dim StringDesc As USB_STRING_DESCRIPTOR = DirectCast(Marshal.PtrToStructure(ptrStringDesc, GetType(USB_STRING_DESCRIPTOR)), USB_STRING_DESCRIPTOR)
                        Device.DeviceSerialNumber = StringDesc.bString
                    End If
                    Marshal.FreeHGlobal(ptrRequest)
                End If

                ' Get the Driver Key Name (usefull in locating a device)
                Dim DriverKey As New USB_NODE_CONNECTION_DRIVERKEY_NAME()
                DriverKey.ConnectionIndex = PortPortNumber
                nBytes = Marshal.SizeOf(DriverKey)
                Dim ptrDriverKey As IntPtr = Marshal.AllocHGlobal(nBytes)
                Marshal.StructureToPtr(DriverKey, ptrDriverKey, True)

                ' Use an IOCTL call to request the Driver Key Name
                If DeviceIoControl(h, IOCTL_USB_GET_NODE_CONNECTION_DRIVERKEY_NAME, ptrDriverKey, nBytes, ptrDriverKey, nBytes, _
                 nBytesReturned, IntPtr.Zero) Then
                    DriverKey = DirectCast(Marshal.PtrToStructure(ptrDriverKey, GetType(USB_NODE_CONNECTION_DRIVERKEY_NAME)), USB_NODE_CONNECTION_DRIVERKEY_NAME)
                    Device.DeviceDriverKey = DriverKey.DriverKeyName

                    ' use the DriverKeyName to get the Device Description and Instance ID
                    Device.DeviceName = GetDescriptionByKeyName(Device.DeviceDriverKey)
                    Device.DeviceInstanceID = GetInstanceIDByKeyName(Device.DeviceDriverKey)
                End If
                Marshal.FreeHGlobal(ptrDriverKey)
                CloseHandle(h)
            End If
            Return Device
        End Function

        '
        ' private function for finding a USB device's Description
        '
        Private Shared Function GetDescriptionByKeyName(ByVal DriverKeyName As String) As String
            Dim ans As String = ""
            Dim DevEnum As String = REGSTR_KEY_USB

            ' Use the "enumerator form" of the SetupDiGetClassDevs API 
            ' to generate a list of all USB devices
            Dim h As IntPtr = SetupDiGetClassDevs(0, DevEnum, IntPtr.Zero, DIGCF_PRESENT Or DIGCF_ALLCLASSES)
            If h.ToInt32() <> INVALID_HANDLE_VALUE Then
                Dim ptrBuf As IntPtr = Marshal.AllocHGlobal(BUFFER_SIZE)
                Dim KeyName As String

                Dim Success As Boolean
                Dim i As Integer = 0
                Do
                    ' create a Device Interface Data structure
                    Dim da As New SP_DEVINFO_DATA()
                    da.cbSize = Marshal.SizeOf(da)

                    ' start the enumeration 
                    Success = SetupDiEnumDeviceInfo(h, i, da)
                    If Success Then
                        Dim RequiredSize As Integer = 0
                        Dim RegType As Integer = REG_SZ
                        KeyName = ""

                        If SetupDiGetDeviceRegistryProperty(h, da, SPDRP_DRIVER, RegType, ptrBuf, BUFFER_SIZE, _
                         RequiredSize) Then
                            KeyName = Marshal.PtrToStringAuto(ptrBuf)
                        End If

                        ' is it a match?
                        If KeyName = DriverKeyName Then
                            If SetupDiGetDeviceRegistryProperty(h, da, SPDRP_DEVICEDESC, RegType, ptrBuf, BUFFER_SIZE, _
                             RequiredSize) Then
                                ans = Marshal.PtrToStringAuto(ptrBuf)
                            End If
                            Exit Do
                        End If
                    End If
                    i += 1
                Loop While Success

                Marshal.FreeHGlobal(ptrBuf)
                SetupDiDestroyDeviceInfoList(h)
            End If
            Return ans
        End Function

        '
        ' private function for finding a USB device's Instance ID
        '
        Private Shared Function GetInstanceIDByKeyName(ByVal DriverKeyName As String) As String
            Dim ans As String = ""
            Dim DevEnum As String = REGSTR_KEY_USB

            ' Use the "enumerator form" of the SetupDiGetClassDevs API 
            ' to generate a list of all USB devices
            Dim h As IntPtr = SetupDiGetClassDevs(0, DevEnum, IntPtr.Zero, DIGCF_PRESENT Or DIGCF_ALLCLASSES)
            If h.ToInt32() <> INVALID_HANDLE_VALUE Then
                Dim ptrBuf As IntPtr = Marshal.AllocHGlobal(BUFFER_SIZE)
                Dim KeyName As String

                Dim Success As Boolean
                Dim i As Integer = 0
                Do
                    ' create a Device Interface Data structure
                    Dim da As New SP_DEVINFO_DATA()
                    da.cbSize = Marshal.SizeOf(da)

                    ' start the enumeration 
                    Success = SetupDiEnumDeviceInfo(h, i, da)
                    If Success Then
                        Dim RequiredSize As Integer = 0
                        Dim RegType As Integer = REG_SZ

                        KeyName = ""
                        If SetupDiGetDeviceRegistryProperty(h, da, SPDRP_DRIVER, RegType, ptrBuf, BUFFER_SIZE, _
                         RequiredSize) Then
                            KeyName = Marshal.PtrToStringAuto(ptrBuf)
                        End If

                        ' is it a match?
                        If KeyName = DriverKeyName Then
                            Dim nBytes As Integer = BUFFER_SIZE
                            Dim sb As New StringBuilder(nBytes)
                            SetupDiGetDeviceInstanceId(h, da, sb, nBytes, RequiredSize)
                            ans = sb.ToString()
                            Exit Do
                        End If
                    End If
                    i += 1
                Loop While Success

                Marshal.FreeHGlobal(ptrBuf)
                SetupDiDestroyDeviceInfoList(h)
            End If
            Return ans
        End Function

        ' return a down stream external hub
        Public Function GetHub() As USBHub
            If Not PortIsHub Then
                Return Nothing
            End If
            Dim Hub As New USBHub()
            Dim h As IntPtr, h2 As IntPtr
            Hub.HubIsRootHub = False
            Hub.HubDeviceDesc = "External Hub"

            ' Open a handle to the Host Controller
            h = CreateFile(PortHubDevicePath, GENERIC_WRITE, FILE_SHARE_WRITE, IntPtr.Zero, OPEN_EXISTING, 0, _
             IntPtr.Zero)
            If h.ToInt32() <> INVALID_HANDLE_VALUE Then
                ' Get the DevicePath for downstream hub
                Dim nBytesReturned As Integer
                Dim NodeName As New USB_NODE_CONNECTION_NAME()
                NodeName.ConnectionIndex = PortPortNumber
                Dim nBytes As Integer = Marshal.SizeOf(NodeName)
                Dim ptrNodeName As IntPtr = Marshal.AllocHGlobal(nBytes)
                Marshal.StructureToPtr(NodeName, ptrNodeName, True)

                ' Use an IOCTL call to request the Node Name
                If DeviceIoControl(h, IOCTL_USB_GET_NODE_CONNECTION_NAME, ptrNodeName, nBytes, ptrNodeName, nBytes, _
                 nBytesReturned, IntPtr.Zero) Then
                    NodeName = DirectCast(Marshal.PtrToStructure(ptrNodeName, GetType(USB_NODE_CONNECTION_NAME)), USB_NODE_CONNECTION_NAME)
                    Hub.HubDevicePath = "\\.\" + NodeName.NodeName
                End If

                ' Now let's open the Hub (based upon the HubName we got above)
                h2 = CreateFile(Hub.HubDevicePath, GENERIC_WRITE, FILE_SHARE_WRITE, IntPtr.Zero, OPEN_EXISTING, 0, _
                 IntPtr.Zero)
                If h2.ToInt32() <> INVALID_HANDLE_VALUE Then
                    Dim NodeInfo As New USB_NODE_INFORMATION()
                    NodeInfo.NodeType = CInt(USB_HUB_NODE.UsbHub)
                    nBytes = Marshal.SizeOf(NodeInfo)
                    Dim ptrNodeInfo As IntPtr = Marshal.AllocHGlobal(nBytes)
                    Marshal.StructureToPtr(NodeInfo, ptrNodeInfo, True)

                    ' get the Hub Information
                    If DeviceIoControl(h2, IOCTL_USB_GET_NODE_INFORMATION, ptrNodeInfo, nBytes, ptrNodeInfo, nBytes, _
                     nBytesReturned, IntPtr.Zero) Then
                        NodeInfo = DirectCast(Marshal.PtrToStructure(ptrNodeInfo, GetType(USB_NODE_INFORMATION)), USB_NODE_INFORMATION)
                        Hub.HubIsBusPowered = Convert.ToBoolean(NodeInfo.HubInformation.HubIsBusPowered)
                        Hub.HubPortCount = NodeInfo.HubInformation.HubDescriptor.bNumberOfPorts
                    End If
                    Marshal.FreeHGlobal(ptrNodeInfo)
                    CloseHandle(h2)
                End If

                ' Fill in the missing Manufacture, Product, and SerialNumber values
                ' values by just creating a Device instance and copying the values
                Dim Device As USBDevice = GetDevice()
                Hub.HubInstanceID = Device.DeviceInstanceID
                Hub.HubManufacturer = Device.Manufacturer
                Hub.HubProduct = Device.Product
                Hub.HubSerialNumber = Device.SerialNumber
                Hub.HubDriverKey = Device.DriverKey

                Marshal.FreeHGlobal(ptrNodeName)
                CloseHandle(h)
            End If
            Return Hub
        End Function



    End Class

End Class
